"""Unit tests for the deck_conflict module."""

from contextlib import nullcontext as does_not_raise
from typing import Any, ContextManager, List, Literal, NamedTuple, Tuple, cast

import pytest
from decoy import Decoy, matchers

from opentrons_shared_data.labware.types import LabwareUri
from opentrons_shared_data.robot.types import RobotType

from ... import versions_at_or_above, versions_below
from opentrons.hardware_control import CriticalPoint
from opentrons.motion_planning import adjacent_slots_getters
from opentrons.motion_planning import deck_conflict as wrapped_deck_conflict
from opentrons.motion_planning.adjacent_slots_getters import _MixedTypeSlots
from opentrons.protocol_api import MAX_SUPPORTED_VERSION
from opentrons.protocol_api.core.engine import deck_conflict, pipette_movement_conflict
from opentrons.protocol_api.disposal_locations import (
    _TRASH_BIN_CUTOUT_FIXTURE,
    TrashBin,
    WasteChute,
)
from opentrons.protocol_api.labware import Labware
from opentrons.protocol_engine import (
    Config,
    DeckSlotLocation,
    ModuleModel,
    StateView,
)
from opentrons.protocol_engine.clients import SyncClient
from opentrons.protocol_engine.errors import LabwareNotLoadedOnModuleError
from opentrons.protocol_engine.state.geometry import _AbsoluteRobotExtents
from opentrons.protocol_engine.state.pipettes import PipetteBoundingBoxOffsets
from opentrons.protocol_engine.types import (
    DeckType,
    Dimensions,
    DropTipWellLocation,
    LoadedLabware,
    LoadedModule,
    OnDeckLabwareLocation,
    OnLabwareLocation,
    StagingSlotLocation,
    WellLocation,
    WellOffset,
    WellOrigin,
)
from opentrons.protocols.api_support.types import APIVersion
from opentrons.types import (
    DeckSlotName,
    MountType,
    NozzleConfigurationType,
    Point,
    StagingSlotName,
)


@pytest.fixture(autouse=True)
def patch_slot_getters(decoy: Decoy, monkeypatch: pytest.MonkeyPatch) -> None:
    """Mock out adjacent_slots_getters functions."""
    mock_get_surrounding_slots = decoy.mock(
        func=adjacent_slots_getters.get_surrounding_slots
    )
    mock_get_surrounding_staging_slots = decoy.mock(
        func=adjacent_slots_getters.get_surrounding_staging_slots
    )
    monkeypatch.setattr(
        adjacent_slots_getters, "get_surrounding_slots", mock_get_surrounding_slots
    )
    monkeypatch.setattr(
        adjacent_slots_getters,
        "get_surrounding_staging_slots",
        mock_get_surrounding_staging_slots,
    )


@pytest.fixture(autouse=True)
def use_mock_wrapped_deck_conflict(
    decoy: Decoy, monkeypatch: pytest.MonkeyPatch
) -> None:
    """Replace the check() function that our subject should wrap with a mock."""
    mock_check = decoy.mock(func=wrapped_deck_conflict.check)
    monkeypatch.setattr(wrapped_deck_conflict, "check", mock_check)


@pytest.fixture
def api_version() -> APIVersion:
    """Get mocked api_version."""
    return MAX_SUPPORTED_VERSION


@pytest.fixture
def mock_sync_client(decoy: Decoy) -> SyncClient:
    """Return a mock in the shape of a SyncClient."""
    return decoy.mock(cls=SyncClient)


@pytest.fixture
def mock_state_view(
    decoy: Decoy,
    robot_type: RobotType,
    deck_type: DeckType,
) -> StateView:
    """Return a mock in the shape of a StateView."""
    mock_state_view = decoy.mock(cls=StateView)
    config = Config(robot_type=robot_type, deck_type=deck_type)
    decoy.when(mock_state_view.config).then_return(config)
    return mock_state_view


@pytest.mark.parametrize(
    ("robot_type", "deck_type"),
    [
        ("OT-2 Standard", DeckType.OT2_STANDARD),
        ("OT-3 Standard", DeckType.OT3_STANDARD),
    ],
)
def test_maps_labware_on_deck(decoy: Decoy, mock_state_view: StateView) -> None:
    """It should correcly map a labware that's loaded directly into a deck slot."""
    decoy.when(
        mock_state_view.labware.get_location(labware_id="labware-id")
    ).then_return(DeckSlotLocation(slotName=DeckSlotName.SLOT_5))

    decoy.when(
        mock_state_view.labware.get_load_name(labware_id="labware-id")
    ).then_return("labware_load_name")
    decoy.when(
        mock_state_view.geometry.get_labware_highest_z(labware_id="labware-id")
    ).then_return(3.14159)
    decoy.when(
        mock_state_view.labware.get_definition_uri(labware_id="labware-id")
    ).then_return(LabwareUri("test/labware_load_name/123"))
    decoy.when(
        mock_state_view.labware.is_fixed_trash(labware_id="labware-id")
    ).then_return(True)

    # Test both ways that the subject's caller can provide a labware:
    # by new_labware_id, and by existing_labware_ids.
    # We reuse the same ID for test convenience; in the real world, they'd be different.
    deck_conflict.check(
        engine_state=mock_state_view,
        existing_labware_ids=["labware-id"],
        existing_module_ids=[],
        existing_disposal_locations=[],
        new_labware_id="labware-id",
    )
    decoy.verify(
        wrapped_deck_conflict.check(
            existing_items={
                DeckSlotName.SLOT_5: wrapped_deck_conflict.Labware(
                    name_for_errors="labware_load_name",
                    highest_z=3.14159,
                    uri=LabwareUri("test/labware_load_name/123"),
                    is_fixed_trash=True,
                )
            },
            new_item=wrapped_deck_conflict.Labware(
                name_for_errors="labware_load_name",
                highest_z=3.14159,
                uri=LabwareUri("test/labware_load_name/123"),
                is_fixed_trash=True,
            ),
            new_location=DeckSlotName.SLOT_5,
            robot_type=mock_state_view.config.robot_type,
        )
    )


@pytest.mark.parametrize(
    ("robot_type", "deck_type"),
    [
        ("OT-2 Standard", DeckType.OT2_STANDARD),
        ("OT-3 Standard", DeckType.OT3_STANDARD),
    ],
)
def test_maps_module_without_labware(decoy: Decoy, mock_state_view: StateView) -> None:
    """It should correctly map a module with no labware loaded atop it."""
    decoy.when(mock_state_view.labware.get_id_by_module("module-id")).then_raise(
        LabwareNotLoadedOnModuleError()
    )
    decoy.when(mock_state_view.modules.get_overall_height("module-id")).then_return(
        3.14159
    )

    decoy.when(mock_state_view.modules.get_connected_model("module-id")).then_return(
        ModuleModel.HEATER_SHAKER_MODULE_V1
    )
    decoy.when(mock_state_view.modules.get_location("module-id")).then_return(
        DeckSlotLocation(slotName=DeckSlotName.SLOT_5)
    )

    # Test both ways that the subject's caller can provide a module:
    # by new_module_id, and by existing_module_ids.
    # We reuse the same ID for test convenience; in the real world, they'd be different.
    deck_conflict.check(
        engine_state=mock_state_view,
        existing_labware_ids=[],
        existing_module_ids=["module-id"],
        existing_disposal_locations=[],
        new_module_id="module-id",
    )
    decoy.verify(
        wrapped_deck_conflict.check(
            existing_items={
                DeckSlotName.SLOT_5: wrapped_deck_conflict.HeaterShakerModule(
                    name_for_errors="heaterShakerModuleV1",
                    highest_z_including_labware=3.14159,
                )
            },
            new_item=wrapped_deck_conflict.HeaterShakerModule(
                name_for_errors="heaterShakerModuleV1",
                highest_z_including_labware=3.14159,
            ),
            new_location=DeckSlotName.SLOT_5,
            robot_type=mock_state_view.config.robot_type,
        )
    )


@pytest.mark.parametrize(
    ("robot_type", "deck_type"),
    [
        ("OT-2 Standard", DeckType.OT2_STANDARD),
        ("OT-3 Standard", DeckType.OT3_STANDARD),
    ],
)
def test_maps_module_with_labware(decoy: Decoy, mock_state_view: StateView) -> None:
    """It should correctly map a module with a labware loaded atop it.

    The highest_z should include both the labware and the module.
    """
    decoy.when(mock_state_view.labware.get_id_by_module("module-id")).then_return(
        "labware-id"
    )
    decoy.when(
        mock_state_view.geometry.get_labware_highest_z("labware-id")
    ).then_return(3.14159)

    decoy.when(mock_state_view.modules.get_connected_model("module-id")).then_return(
        ModuleModel.HEATER_SHAKER_MODULE_V1
    )
    decoy.when(mock_state_view.modules.get_location("module-id")).then_return(
        DeckSlotLocation(slotName=DeckSlotName.SLOT_5)
    )

    # Test both ways that the subject's caller can provide a module:
    # by new_module_id, and by existing_module_ids.
    # We reuse the same ID for test convenience; in the real world, they'd be different.
    deck_conflict.check(
        engine_state=mock_state_view,
        existing_labware_ids=[],
        existing_module_ids=["module-id"],
        existing_disposal_locations=[],
        new_module_id="module-id",
    )
    decoy.verify(
        wrapped_deck_conflict.check(
            existing_items={
                DeckSlotName.SLOT_5: wrapped_deck_conflict.HeaterShakerModule(
                    name_for_errors="heaterShakerModuleV1",
                    highest_z_including_labware=3.14159,
                )
            },
            new_item=wrapped_deck_conflict.HeaterShakerModule(
                name_for_errors="heaterShakerModuleV1",
                highest_z_including_labware=3.14159,
            ),
            new_location=DeckSlotName.SLOT_5,
            robot_type=mock_state_view.config.robot_type,
        )
    )


@pytest.mark.parametrize(
    ("robot_type", "deck_type"),
    [
        ("OT-2 Standard", DeckType.OT2_STANDARD),
        ("OT-3 Standard", DeckType.OT3_STANDARD),
    ],
)
@pytest.mark.parametrize("module_model", ModuleModel)
def test_maps_different_module_models(
    decoy: Decoy, mock_state_view: StateView, module_model: ModuleModel
) -> None:
    """It should correctly map all possible kinds of hardware module."""

    def get_expected_mapping_result() -> wrapped_deck_conflict.DeckItem:
        expected_name_for_errors = module_model.value
        if module_model is ModuleModel.HEATER_SHAKER_MODULE_V1:
            return wrapped_deck_conflict.HeaterShakerModule(
                name_for_errors=expected_name_for_errors,
                highest_z_including_labware=3.14159,
            )
        elif module_model is ModuleModel.MAGNETIC_BLOCK_V1:
            return wrapped_deck_conflict.MagneticBlockModule(
                name_for_errors=expected_name_for_errors,
                highest_z_including_labware=3.14159,
            )
        elif (
            module_model is ModuleModel.THERMOCYCLER_MODULE_V1
            or module_model is ModuleModel.THERMOCYCLER_MODULE_V2
        ):
            return wrapped_deck_conflict.ThermocyclerModule(
                name_for_errors=expected_name_for_errors,
                highest_z_including_labware=3.14159,
                is_semi_configuration=False,
            )
        elif module_model is ModuleModel.FLEX_STACKER_MODULE_V1:
            return wrapped_deck_conflict.FlexStackerModule(
                name_for_errors=expected_name_for_errors,
                highest_z_including_labware=3.14159,
            )
        else:
            return wrapped_deck_conflict.OtherModule(
                name_for_errors=expected_name_for_errors,
                highest_z_including_labware=3.14159,
            )
        # There is deliberately no catch-all `else` block here.
        # If a new value is added to ModuleModel, it should cause an error here and
        # force us to think about how it should be mapped.

    decoy.when(mock_state_view.modules.get_connected_model("module-id")).then_return(
        module_model
    )

    decoy.when(mock_state_view.labware.get_id_by_module("module-id")).then_raise(
        LabwareNotLoadedOnModuleError()
    )
    decoy.when(mock_state_view.modules.get_overall_height("module-id")).then_return(
        3.14159
    )
    decoy.when(mock_state_view.modules.get_location("module-id")).then_return(
        DeckSlotLocation(slotName=DeckSlotName.SLOT_5)
    )

    expected_mapping_result = get_expected_mapping_result()

    deck_conflict.check(
        engine_state=mock_state_view,
        existing_labware_ids=[],
        existing_module_ids=[],
        existing_disposal_locations=[],
        new_module_id="module-id",
    )
    decoy.verify(
        wrapped_deck_conflict.check(
            existing_items={},
            new_item=expected_mapping_result,
            new_location=DeckSlotName.SLOT_5,
            robot_type=mock_state_view.config.robot_type,
        )
    )


@pytest.mark.parametrize(
    ("robot_type", "deck_type"),
    [
        ("OT-2 Standard", DeckType.OT2_STANDARD),
        ("OT-3 Standard", DeckType.OT3_STANDARD),
    ],
)
def test_maps_trash_bins(
    decoy: Decoy,
    mock_state_view: StateView,
    api_version: APIVersion,
    mock_sync_client: SyncClient,
) -> None:
    """It should correctly map disposal locations."""
    mock_trash_lw = decoy.mock(cls=Labware)

    decoy.when(
        mock_sync_client.state.addressable_areas.get_fixture_height(
            _TRASH_BIN_CUTOUT_FIXTURE
        )
    ).then_return(1.23)

    deck_conflict.check(
        engine_state=mock_state_view,
        existing_labware_ids=[],
        existing_module_ids=[],
        existing_disposal_locations=[
            TrashBin(
                location=DeckSlotName.SLOT_B1,
                addressable_area_name="blah",
                engine_client=mock_sync_client,
                api_version=api_version,
            ),
            WasteChute(engine_client=mock_sync_client, api_version=api_version),
            mock_trash_lw,
        ],
        new_trash_bin=TrashBin(
            location=DeckSlotName.SLOT_A1,
            addressable_area_name="blah",
            engine_client=mock_sync_client,
            api_version=api_version,
        ),
    )
    decoy.verify(
        wrapped_deck_conflict.check(
            existing_items={
                DeckSlotName.SLOT_B1: wrapped_deck_conflict.TrashBin(
                    name_for_errors="trash bin", highest_z=1.23
                )
            },
            new_item=wrapped_deck_conflict.TrashBin(
                name_for_errors="trash bin", highest_z=1.23
            ),
            new_location=DeckSlotName.SLOT_A1,
            robot_type=mock_state_view.config.robot_type,
        )
    )


def _modules_non_stacker() -> list[ModuleModel]:
    return [m for m in ModuleModel if not ModuleModel.is_flex_stacker(m)]


_lw_index = 0
_mod_index = 0


def _provide_item_in_state(
    decoy: Decoy,
    mock_state_view: StateView,
    mock_sync_client: SyncClient,
    item_type: Literal["labware", "trash-bin"] | ModuleModel,
) -> tuple[str | None, str | None, TrashBin | None]:
    global _lw_index, _mod_index
    if item_type == "labware":
        labware_id = f"labware-id-{_lw_index}"
        _lw_index += 1
        decoy.when(
            mock_state_view.labware.get_location(labware_id=labware_id)
        ).then_return(DeckSlotLocation(slotName=DeckSlotName.SLOT_5))
        decoy.when(
            mock_state_view.labware.get_load_name(labware_id=labware_id)
        ).then_return("labware_load_name")
        decoy.when(
            mock_state_view.geometry.get_labware_highest_z(labware_id=labware_id)
        ).then_return(3.14159)
        decoy.when(
            mock_state_view.labware.get_definition_uri(labware_id=labware_id)
        ).then_return(LabwareUri("test/labware_load_name/123"))
        decoy.when(
            mock_state_view.labware.is_fixed_trash(labware_id=labware_id)
        ).then_return(False)
        return (labware_id, None, None)
    elif item_type == "trash-bin":
        decoy.when(
            mock_sync_client.state.addressable_areas.get_fixture_height(
                _TRASH_BIN_CUTOUT_FIXTURE
            )
        ).then_return(1.23)
        return (
            None,
            None,
            TrashBin(
                location=DeckSlotName.SLOT_5,
                addressable_area_name="blah",
                engine_client=mock_sync_client,
                api_version=APIVersion(2, 23),
            ),
        )
    else:
        module_id = f"module-id-{_mod_index}"
        _mod_index += 1
        decoy.when(mock_state_view.modules.get_connected_model(module_id)).then_return(
            # this type ignore makes up for a narrowing failure - mypy thinks item_type can
            # still be one of the literals
            item_type  # type: ignore[arg-type]
        )
        decoy.when(mock_state_view.modules.get_location(module_id)).then_return(
            DeckSlotLocation(slotName=DeckSlotName.SLOT_5)
        )
        return (None, module_id, None)


def occupancy_check_state_setup(
    decoy: Decoy,
    mock_state_view: StateView,
    request: pytest.FixtureRequest,
    mock_sync_client: SyncClient,
) -> dict[str, Any]:
    """Set up the current state in the engine occupancy check."""
    kind = cast(Literal["labware", "trash-bin"] | ModuleModel, request.param)
    labware_id, module_id, trash_bin = _provide_item_in_state(
        decoy, mock_state_view, mock_sync_client, kind
    )

    return {
        "existing_labware_ids": [labware_id] if labware_id is not None else [],
        "existing_module_ids": [module_id] if module_id is not None else [],
        "existing_disposal_locations": [trash_bin] if trash_bin is not None else [],
    }


@pytest.fixture
def occupancy_check_state_setup1(
    decoy: Decoy,
    mock_state_view: StateView,
    request: pytest.FixtureRequest,
    mock_sync_client: SyncClient,
) -> dict[str, Any]:
    """First preconfigured item in the occupancy check."""
    return occupancy_check_state_setup(
        decoy, mock_state_view, request, mock_sync_client
    )


@pytest.fixture
def occupancy_check_state_setup2(
    decoy: Decoy,
    mock_state_view: StateView,
    request: pytest.FixtureRequest,
    mock_sync_client: SyncClient,
) -> dict[str, Any]:
    """Second preconfigured item in the occupancy check."""
    return occupancy_check_state_setup(
        decoy, mock_state_view, request, mock_sync_client
    )


@pytest.fixture
def occupancy_check_new_item(
    decoy: Decoy,
    mock_state_view: StateView,
    request: pytest.FixtureRequest,
    mock_sync_client: SyncClient,
) -> dict[str, Any]:
    """Set up the engine for the new-item side for the occupancy check."""
    kind = cast(Literal["labware", "trash-bin"] | ModuleModel, request.param)
    labware_id, module_id, trash_bin = _provide_item_in_state(
        decoy, mock_state_view, mock_sync_client, kind
    )
    return {
        "new_labware_id": labware_id,
        "new_module_id": module_id,
        "new_trash_bin": trash_bin,
    }


@pytest.mark.parametrize(
    "occupancy_check_state_setup1",
    ["labware", "trash-bin"] + _modules_non_stacker(),  # type: ignore[operator]
    indirect=True,
)
@pytest.mark.parametrize(
    "occupancy_check_state_setup2",
    ["labware", "trash-bin"] + _modules_non_stacker(),  # type: ignore[operator]
    indirect=True,
)
@pytest.mark.parametrize(
    "occupancy_check_new_item",
    ["labware"],
    indirect=True,
)
@pytest.mark.parametrize(
    ("robot_type", "deck_type"),
    [
        ("OT-2 Standard", DeckType.OT2_STANDARD),
        ("OT-3 Standard", DeckType.OT3_STANDARD),
    ],
)
def test_maps_default_single_occupancy(
    decoy: Decoy,
    mock_state_view: StateView,
    occupancy_check_state_setup1: dict[str, Any],
    occupancy_check_state_setup2: dict[str, Any],
    occupancy_check_new_item: dict[str, Any],
) -> None:
    """It should correctly prevent double occupancy when a stacker is not involved."""
    setup = {**occupancy_check_state_setup1}
    for k, v in occupancy_check_state_setup2.items():
        setup[k].extend(v)
    with pytest.raises(
        wrapped_deck_conflict.DeckConflictError, match="cannot both be loaded in"
    ):
        deck_conflict.check(
            engine_state=mock_state_view,
            **setup,
            **occupancy_check_new_item,
        )


@pytest.mark.parametrize(
    "occupancy_check_state_setup1",
    [m for m in ModuleModel if ModuleModel.is_flex_stacker(m)],
    indirect=True,
)
@pytest.mark.parametrize(
    "occupancy_check_state_setup2",
    ["labware"] + [m for m in ModuleModel if ModuleModel.is_magnetic_block(m)],
    indirect=True,
)
@pytest.mark.parametrize(
    "occupancy_check_new_item",
    ["labware"],
    indirect=True,
)
@pytest.mark.parametrize(
    ("robot_type", "deck_type"),
    [
        ("OT-3 Standard", DeckType.OT3_STANDARD),
    ],
)
def test_maps_allows_stacker_labware_double_occupancy(
    decoy: Decoy,
    mock_state_view: StateView,
    occupancy_check_state_setup1: dict[str, Any],
    occupancy_check_state_setup2: dict[str, Any],
    occupancy_check_new_item: dict[str, Any],
) -> None:
    """It should correctly allow double occupancy for a stacker and labware or mag block."""
    decoy.when(
        wrapped_deck_conflict.check(
            existing_items=matchers.Anything(),
            new_item=matchers.Anything(),
            new_location=matchers.Anything(),
            robot_type=matchers.Anything(),
        )
    ).then_return(True)
    setup_a = {k: [i for i in v] for k, v in occupancy_check_state_setup1.items()}
    for k, v in occupancy_check_state_setup2.items():
        setup_a[k].extend(v)
    deck_conflict.check(  # type: ignore[call-overload]
        engine_state=mock_state_view,
        **setup_a,
        **occupancy_check_new_item,
    )
    setup_b = {k: [i for i in v] for k, v in occupancy_check_state_setup2.items()}
    for k, v in occupancy_check_state_setup1.items():
        setup_b[k].extend(v)
    deck_conflict.check(  # type: ignore[call-overload]
        engine_state=mock_state_view,
        **setup_b,
        **occupancy_check_new_item,
    )


@pytest.mark.parametrize(
    "occupancy_check_state_setup1",
    [m for m in ModuleModel if ModuleModel.is_flex_stacker(m)],
    indirect=True,
)
@pytest.mark.parametrize(
    "occupancy_check_state_setup2",
    ["trash-bin"] + [m for m in ModuleModel if not ModuleModel.is_magnetic_block(m)],
    indirect=True,
)
@pytest.mark.parametrize(
    "occupancy_check_new_item",
    ["labware"],
    indirect=True,
)
@pytest.mark.parametrize(
    ("robot_type", "deck_type"),
    [
        ("OT-3 Standard", DeckType.OT3_STANDARD),
    ],
)
def test_maps_prevents_stacker_non_labware_double_occupancy(
    decoy: Decoy,
    mock_state_view: StateView,
    occupancy_check_state_setup1: dict[str, Any],
    occupancy_check_state_setup2: dict[str, Any],
    occupancy_check_new_item: dict[str, Any],
) -> None:
    """It should correctly allow double occupancy for a stacker and labware or mag block."""
    setup = {**occupancy_check_state_setup1}
    for k, v in occupancy_check_state_setup2.items():
        setup[k].extend(v)
    with pytest.raises(
        wrapped_deck_conflict.DeckConflictError, match="cannot both be loaded in"
    ):
        deck_conflict.check(
            engine_state=mock_state_view,
            **setup,
            **occupancy_check_new_item,
        )


plate = LoadedLabware(
    id="plate-id",
    loadName="plate-load-name",
    location=DeckSlotLocation(slotName=DeckSlotName.SLOT_C1),
    definitionUri="some-plate-uri",
    offsetId=None,
    displayName="Fancy Plate Name",
)

module = LoadedModule(
    id="module-id",
    model=ModuleModel.TEMPERATURE_MODULE_V1,
    location=DeckSlotLocation(slotName=DeckSlotName.SLOT_C1),
    serialNumber="serial-number",
)


@pytest.mark.parametrize(
    ("robot_type", "deck_type"),
    [("OT-3 Standard", DeckType.OT3_STANDARD)],
)
@pytest.mark.parametrize(
    ["pipette_bounds", "expected_raise", "y_value"],
    [
        (  # nozzles above highest Z
            (
                Point(x=50, y=150, z=60),
                Point(x=150, y=50, z=60),
                Point(x=150, y=150, z=60),
                Point(x=50, y=50, z=60),
            ),
            does_not_raise(),
            0,
        ),
        # X, Y, Z collisions
        (
            (
                Point(x=50, y=150, z=40),
                Point(x=150, y=50, z=40),
                Point(x=150, y=150, z=40),
                Point(x=50, y=50, z=40),
            ),
            pytest.raises(
                pipette_movement_conflict.PartialTipMovementNotAllowedError,
                match="collision with items in deck slot D1",
            ),
            0,
        ),
        (
            (
                Point(x=101, y=150, z=40),
                Point(x=150, y=50, z=40),
                Point(x=150, y=150, z=40),
                Point(x=101, y=50, z=40),
            ),
            pytest.raises(
                pipette_movement_conflict.PartialTipMovementNotAllowedError,
                match="collision with items in deck slot D2",
            ),
            0,
        ),
        (  # Collision with staging slot
            (
                Point(x=150, y=150, z=40),
                Point(x=250, y=101, z=40),
                Point(x=150, y=101, z=40),
                Point(x=250, y=150, z=40),
            ),
            pytest.raises(
                pipette_movement_conflict.PartialTipMovementNotAllowedError,
                match="will result in collision with items in staging slot C4.",
            ),
            170,
        ),
        (
            (
                Point(x=150, y=250, z=40),
                Point(x=250, y=201, z=40),
                Point(x=150, y=201, z=40),
                Point(x=250, y=250, z=40),
            ),
            pytest.raises(
                pipette_movement_conflict.PartialTipMovementNotAllowedError,
                match="result in collision with items on flexStackerModuleV1 mounted in B3.",
            ),
            170,
        ),
    ],
)
def test_deck_conflict_raises_for_bad_pipette_move(
    decoy: Decoy,
    mock_state_view: StateView,
    api_version: APIVersion,
    pipette_bounds: Tuple[Point, Point, Point, Point],
    expected_raise: ContextManager[Any],
    y_value: float,
) -> None:
    """It should raise errors when moving to locations with restrictions for partial pipette movement.

    Test premise:
    - we are using a pipette configured for COLUMN nozzle layout with primary nozzle A12
    - there are labware of height 50mm in C1, D1 & D2
    - we are checking for conflicts when moving to a labware in C2.
      For each test case, we are moving to a different point in the destination labware,
      with the same pipette and tip
    - we are checking for conflicts when moving to a point that would collide with a
      flex stacker in column 4 at position B4 but nothing in the ancestor slot of B3

    Note: this test does not stub out the slot overlap checker function
          in order to preserve readability of the test. That means the test does
          actual slot overlap checks.
    """
    destination_well_point = Point(x=123, y=y_value, z=123)
    decoy.when(
        mock_state_view.pipettes.get_is_partially_configured("pipette-id")
    ).then_return(True)
    decoy.when(mock_state_view.pipettes.get_mount("pipette-id")).then_return(
        MountType.LEFT
    )
    decoy.when(mock_state_view.geometry.absolute_deck_extents).then_return(
        _AbsoluteRobotExtents(
            front_left={
                MountType.LEFT: Point(13.5, -60.5, 0.0),
                MountType.RIGHT: Point(-40.5, -60.5, 0.0),
            },
            back_right={
                MountType.LEFT: Point(463.7, 433.3, 0.0),
                MountType.RIGHT: Point(517.7, 433.3),
            },
            deck_extents=Point(477.2, 493.8, 0.0),
            padding_rear=-181.21,
            padding_front=55.8,
            padding_left_side=31.88,
            padding_right_side=-80.32,
        )
    )
    decoy.when(
        mock_state_view.pipettes.get_pipette_bounding_box("pipette-id")
    ).then_return(
        # 96 chan outer bounds
        PipetteBoundingBoxOffsets(
            back_left_corner=Point(-36.0, -25.5, -259.15),
            front_right_corner=Point(63.0, -88.5, -259.15),
            front_left_corner=Point(-36.0, -88.5, -259.15),
            back_right_corner=Point(63.0, -25.5, -259.15),
        )
    )
    decoy.when(
        mock_state_view.pipettes.get_is_partially_configured("pipette-id")
    ).then_return(True)
    decoy.when(mock_state_view.pipettes.get_primary_nozzle("pipette-id")).then_return(
        "A12"
    )
    decoy.when(
        mock_state_view.geometry.get_ancestor_slot_name("destination-labware-id")
    ).then_return(DeckSlotName.SLOT_C2)

    decoy.when(
        mock_state_view.geometry.get_well_position(
            labware_id="destination-labware-id",
            well_name="A2",
            well_location=WellLocation(origin=WellOrigin.TOP, offset=WellOffset(z=10)),
            pipette_id="pipette-id",
        )
    ).then_return(destination_well_point)
    decoy.when(
        mock_state_view.motion.get_critical_point_for_wells_in_labware(
            "destination-labware-id"
        )
    ).then_return(None)
    decoy.when(
        mock_state_view.pipettes.get_pipette_bounds_at_specified_move_to_position(
            pipette_id="pipette-id",
            destination_position=destination_well_point,
            critical_point=None,
        )
    ).then_return(pipette_bounds)

    stacker = LoadedModule(
        id="fake-stacker-id",
        model=ModuleModel.FLEX_STACKER_MODULE_V1,
        location=DeckSlotLocation(slotName=DeckSlotName.SLOT_B3),
        serialNumber="serial-number",
    )
    decoy.when(mock_state_view.modules.get_by_slot(DeckSlotName.SLOT_B3)).then_return(
        stacker
    )
    decoy.when(mock_state_view.modules.get_by_slot(DeckSlotName.SLOT_C2)).then_return(
        None
    )
    decoy.when(mock_state_view.modules.is_column_4_module(stacker.model)).then_return(
        True
    )
    decoy.when(
        mock_state_view.modules.ensure_and_convert_module_fixture_location(
            DeckSlotName.SLOT_B3, stacker.model
        )
    ).then_return("flexStackerModuleV1B4")

    decoy.when(
        adjacent_slots_getters.get_surrounding_slots(5, robot_type="OT-3 Standard")
    ).then_return(
        _MixedTypeSlots(
            regular_slots=[
                DeckSlotName.SLOT_D1,
                DeckSlotName.SLOT_D2,
                DeckSlotName.SLOT_C1,
                DeckSlotName.SLOT_B3,
            ],
            staging_slots=[StagingSlotName.SLOT_C4],
        )
    )
    decoy.when(
        adjacent_slots_getters.get_surrounding_staging_slots(DeckSlotName.SLOT_C2)
    ).then_return([StagingSlotName.SLOT_C4])

    decoy.when(
        mock_state_view.addressable_areas.get_addressable_area_position(
            addressable_area_name="C1", do_compatibility_check=False
        )
    ).then_return(Point(0, 100, 0))
    decoy.when(
        mock_state_view.addressable_areas.get_addressable_area_position(
            addressable_area_name="D1", do_compatibility_check=False
        )
    ).then_return(Point(0, 0, 0))
    decoy.when(
        mock_state_view.addressable_areas.get_addressable_area_position(
            addressable_area_name="D2", do_compatibility_check=False
        )
    ).then_return(Point(100, 0, 0))
    decoy.when(
        mock_state_view.addressable_areas.get_addressable_area_position(
            addressable_area_name="C4", do_compatibility_check=False
        )
    ).then_return(Point(200, 100, 0))
    decoy.when(
        mock_state_view.addressable_areas.get_addressable_area_bounding_box(
            addressable_area_name="C4", do_compatibility_check=False
        )
    ).then_return(Dimensions(90, 90, 0))
    decoy.when(
        mock_state_view.geometry.get_highest_z_in_slot(
            StagingSlotLocation(slotName=StagingSlotName.SLOT_C4)
        )
    ).then_return(50)
    decoy.when(
        mock_state_view.addressable_areas.get_addressable_area_position(
            addressable_area_name="B3", do_compatibility_check=False
        )
    ).then_return(Point(150, 200, 0))
    decoy.when(
        mock_state_view.addressable_areas.get_addressable_area_bounding_box(
            addressable_area_name="B3", do_compatibility_check=False
        )
    ).then_return(Dimensions(90, 90, 0))

    # Ensure slot B3 is empty so we can test the stacker
    decoy.when(
        mock_state_view.geometry.get_highest_z_in_slot(
            DeckSlotLocation(slotName=DeckSlotName.SLOT_B3)
        )
    ).then_return(0)

    decoy.when(
        mock_state_view.addressable_areas.get_addressable_area_position(
            addressable_area_name="flexStackerModuleV1B4", do_compatibility_check=False
        )
    ).then_return(Point(200, 200, 0))
    decoy.when(
        mock_state_view.addressable_areas.get_addressable_area_bounding_box(
            addressable_area_name="flexStackerModuleV1B4", do_compatibility_check=False
        )
    ).then_return(Dimensions(90, 90, 0))

    decoy.when(
        mock_state_view.geometry.get_highest_z_of_column_4_module(stacker)
    ).then_return(50)
    for slot_name in [DeckSlotName.SLOT_C1, DeckSlotName.SLOT_D1, DeckSlotName.SLOT_D2]:
        decoy.when(
            mock_state_view.geometry.get_highest_z_in_slot(
                DeckSlotLocation(slotName=slot_name)
            )
        ).then_return(50)
        decoy.when(
            mock_state_view.addressable_areas.get_addressable_area_bounding_box(
                addressable_area_name=slot_name.id, do_compatibility_check=False
            )
        ).then_return(Dimensions(90, 90, 0))

    with expected_raise:
        pipette_movement_conflict.check_safe_for_pipette_movement(
            engine_state=mock_state_view,
            pipette_id="pipette-id",
            labware_id="destination-labware-id",
            well_name="A2",
            well_location=WellLocation(origin=WellOrigin.TOP, offset=WellOffset(z=10)),
            version=api_version,
        )


@pytest.mark.parametrize(
    ("robot_type", "deck_type"),
    [("OT-3 Standard", DeckType.OT3_STANDARD)],
)
def test_deck_conflict_raises_for_collision_with_tc_lid(
    decoy: Decoy,
    mock_state_view: StateView,
    api_version: APIVersion,
) -> None:
    """It should raise an error if pipette might collide with thermocycler lid on the Flex."""
    destination_well_point = Point(x=123, y=123, z=123)
    pipette_bounds_at_destination = (
        Point(x=50, y=350, z=204.5),
        Point(x=150, y=429, z=204.5),
        Point(x=150, y=400, z=204.5),
        Point(x=50, y=300, z=204.5),
    )

    decoy.when(
        mock_state_view.pipettes.get_is_partially_configured("pipette-id")
    ).then_return(True)
    decoy.when(mock_state_view.pipettes.get_primary_nozzle("pipette-id")).then_return(
        "A12"
    )
    decoy.when(
        mock_state_view.geometry.get_ancestor_slot_name("destination-labware-id")
    ).then_return(DeckSlotName.SLOT_C2)

    decoy.when(
        mock_state_view.geometry.get_well_position(
            labware_id="destination-labware-id",
            well_name="A2",
            well_location=WellLocation(origin=WellOrigin.TOP, offset=WellOffset(z=10)),
            pipette_id="pipette-id",
        )
    ).then_return(destination_well_point)

    decoy.when(
        mock_state_view.motion.get_critical_point_for_wells_in_labware(
            "destination-labware-id"
        )
    ).then_return(CriticalPoint.Y_CENTER)
    decoy.when(
        mock_state_view.pipettes.get_pipette_bounds_at_specified_move_to_position(
            pipette_id="pipette-id",
            destination_position=destination_well_point,
            critical_point=CriticalPoint.Y_CENTER,
        )
    ).then_return(pipette_bounds_at_destination)
    decoy.when(mock_state_view.pipettes.get_mount("pipette-id")).then_return(
        MountType.LEFT
    )
    decoy.when(
        mock_state_view.pipettes.get_pipette_bounding_box("pipette-id")
    ).then_return(
        # 96 chan outer bounds
        PipetteBoundingBoxOffsets(
            back_left_corner=Point(-67.0, -3.5, -259.15),
            front_right_corner=Point(94.0, -113.0, -259.15),
            front_left_corner=Point(-67.0, -113.0, -259.15),
            back_right_corner=Point(94.0, -3.5, -259.15),
        )
    )
    decoy.when(mock_state_view.geometry.absolute_deck_extents).then_return(
        _AbsoluteRobotExtents(
            front_left={
                MountType.LEFT: Point(13.5, 60.5, 0.0),
                MountType.RIGHT: Point(-40.5, 60.5, 0.0),
            },
            back_right={
                MountType.LEFT: Point(463.7, 433.3, 0.0),
                MountType.RIGHT: Point(517.7, 433.3),
            },
            deck_extents=Point(477.2, 493.8, 0.0),
            padding_rear=-181.21,
            padding_front=55.8,
            padding_left_side=31.88,
            padding_right_side=-80.32,
        )
    )

    decoy.when(
        adjacent_slots_getters.get_surrounding_slots(5, robot_type="OT-3 Standard")
    ).then_return(
        _MixedTypeSlots(
            regular_slots=[
                DeckSlotName.SLOT_A1,
                DeckSlotName.SLOT_B1,
            ],
            staging_slots=[StagingSlotName.SLOT_C4],
        )
    )
    decoy.when(mock_state_view.modules.is_flex_deck_with_thermocycler()).then_return(
        True
    )
    with pytest.raises(
        pipette_movement_conflict.PartialTipMovementNotAllowedError,
        match="Requested motion with the A12 nozzle partial configuration is outside of robot bounds for the pipette.",
    ):
        pipette_movement_conflict.check_safe_for_pipette_movement(
            engine_state=mock_state_view,
            pipette_id="pipette-id",
            labware_id="destination-labware-id",
            well_name="A2",
            well_location=WellLocation(origin=WellOrigin.TOP, offset=WellOffset(z=10)),
            version=api_version,
        )


@pytest.mark.parametrize(
    ("robot_type", "deck_type"),
    [("OT-3 Standard", DeckType.OT3_STANDARD)],
)
@pytest.mark.parametrize(
    ["destination_well_point", "expected_raise"],
    [
        (
            Point(x=-12, y=100, z=60),
            pytest.raises(
                deck_conflict.PartialTipMovementNotAllowedError,
                match="outside of robot bounds",
            ),
        ),
        (
            Point(x=593, y=100, z=60),
            pytest.raises(
                deck_conflict.PartialTipMovementNotAllowedError,
                match="outside of robot bounds",
            ),
        ),
        (
            Point(x=100, y=1, z=60),
            pytest.raises(
                deck_conflict.PartialTipMovementNotAllowedError,
                match="outside of robot bounds",
            ),
        ),
        (
            Point(x=100, y=507, z=60),
            pytest.raises(
                deck_conflict.PartialTipMovementNotAllowedError,
                match="outside of robot bounds",
            ),
        ),
    ],
)
def test_deck_conflict_raises_for_out_of_bounds_96_channel_move(
    decoy: Decoy,
    mock_state_view: StateView,
    destination_well_point: Point,
    expected_raise: ContextManager[Any],
) -> None:
    """It should raise errors when moving to locations out of robot's bounds for partial tip 96-channel movement.

    Test premise:
    - we are using a pipette configured for COLUMN nozzle layout with primary nozzle A12
    """
    decoy.when(mock_state_view.pipettes.get_channels("pipette-id")).then_return(96)
    decoy.when(
        mock_state_view.labware.get_display_name("destination-labware-id")
    ).then_return("destination-labware")
    decoy.when(
        mock_state_view.pipettes.get_nozzle_layout_type("pipette-id")
    ).then_return(NozzleConfigurationType.COLUMN)
    decoy.when(
        mock_state_view.pipettes.get_is_partially_configured("pipette-id")
    ).then_return(True)
    decoy.when(mock_state_view.pipettes.get_primary_nozzle("pipette-id")).then_return(
        "A12"
    )
    decoy.when(
        mock_state_view.geometry.get_ancestor_slot_name("destination-labware-id")
    ).then_return(DeckSlotName.SLOT_C2)

    decoy.when(
        mock_state_view.geometry.get_well_position(
            labware_id="destination-labware-id",
            well_name="A2",
            well_location=WellLocation(origin=WellOrigin.TOP, offset=WellOffset(z=10)),
            pipette_id="pipette-id",
        )
    ).then_return(destination_well_point)


class PipetteMovementSpec(NamedTuple):
    """Spec data to test deck_conflict.check_safe_for_tip_pickup_and_return ."""

    tiprack_parent: OnDeckLabwareLocation
    tiprack_dim: Dimensions
    is_on_flex_adapter: bool
    is_partial_config: bool
    expected_raise: ContextManager[Any]


pipette_movement_specs: List[PipetteMovementSpec] = [
    PipetteMovementSpec(
        tiprack_parent=DeckSlotLocation(slotName=DeckSlotName.SLOT_5),
        tiprack_dim=Dimensions(x=0, y=0, z=50),
        is_on_flex_adapter=False,
        is_partial_config=False,
        expected_raise=pytest.raises(
            pipette_movement_conflict.UnsuitableTiprackForPipetteMotion,
            match="A cool tiprack must be on an Opentrons Flex 96 Tip Rack Adapter",
        ),
    ),
    PipetteMovementSpec(
        tiprack_parent=OnLabwareLocation(labwareId="adapter-id"),
        tiprack_dim=Dimensions(x=0, y=0, z=50),
        is_on_flex_adapter=True,
        is_partial_config=False,
        expected_raise=does_not_raise(),
    ),
    PipetteMovementSpec(
        tiprack_parent=OnLabwareLocation(labwareId="adapter-id"),
        tiprack_dim=Dimensions(x=0, y=0, z=50),
        is_on_flex_adapter=False,
        is_partial_config=False,
        expected_raise=pytest.raises(
            pipette_movement_conflict.UnsuitableTiprackForPipetteMotion,
            match="A cool tiprack must be on an Opentrons Flex 96 Tip Rack Adapter",
        ),
    ),
    PipetteMovementSpec(
        tiprack_parent=OnLabwareLocation(labwareId="adapter-id"),
        tiprack_dim=Dimensions(x=0, y=0, z=50),
        is_on_flex_adapter=True,
        is_partial_config=True,
        expected_raise=pytest.raises(
            pipette_movement_conflict.PartialTipMovementNotAllowedError,
            match="A cool tiprack cannot be on an adapter taller than the tip rack",
        ),
    ),
    PipetteMovementSpec(
        tiprack_parent=OnLabwareLocation(labwareId="adapter-id"),
        tiprack_dim=Dimensions(x=0, y=0, z=101),
        is_on_flex_adapter=True,
        is_partial_config=True,
        expected_raise=does_not_raise(),
    ),
    PipetteMovementSpec(
        tiprack_parent=DeckSlotLocation(slotName=DeckSlotName.SLOT_5),
        tiprack_dim=Dimensions(x=0, y=0, z=50),
        is_on_flex_adapter=True,  # will be ignored
        is_partial_config=True,
        expected_raise=does_not_raise(),
    ),
]


@pytest.mark.parametrize(
    ("robot_type", "deck_type"),
    [("OT-3 Standard", DeckType.OT3_STANDARD)],
)
@pytest.mark.parametrize(
    argnames=PipetteMovementSpec._fields,
    argvalues=pipette_movement_specs,
)
def test_valid_96_pipette_movement_for_tiprack_and_adapter(
    decoy: Decoy,
    mock_state_view: StateView,
    tiprack_parent: OnDeckLabwareLocation,
    tiprack_dim: Dimensions,
    is_on_flex_adapter: bool,
    is_partial_config: bool,
    expected_raise: ContextManager[Any],
) -> None:
    """It should raise appropriate error for unsuitable tiprack parent when moving 96 channel to it."""
    decoy.when(mock_state_view.pipettes.get_channels("pipette-id")).then_return(96)
    decoy.when(
        mock_state_view.labware.get_dimensions(labware_id="adapter-id")
    ).then_return(Dimensions(x=0, y=0, z=100))
    decoy.when(mock_state_view.labware.get_display_name("labware-id")).then_return(
        "A cool tiprack"
    )
    decoy.when(
        mock_state_view.pipettes.get_is_partially_configured("pipette-id")
    ).then_return(is_partial_config)
    decoy.when(mock_state_view.labware.get_location("labware-id")).then_return(
        tiprack_parent
    )
    decoy.when(
        mock_state_view.labware.get_dimensions(labware_id="labware-id")
    ).then_return(tiprack_dim)
    decoy.when(
        mock_state_view.labware.get_has_quirk(
            labware_id="adapter-id", quirk="tiprackAdapterFor96Channel"
        )
    ).then_return(is_on_flex_adapter)

    with expected_raise:
        pipette_movement_conflict.check_safe_for_tip_pickup_and_return(
            engine_state=mock_state_view,
            pipette_id="pipette-id",
            labware_id="labware-id",
        )


@pytest.mark.parametrize(
    ("robot_type", "deck_type"),
    [
        ("OT-3 Standard", DeckType.OT3_STANDARD),
    ],
)
@pytest.mark.parametrize("api_version", versions_at_or_above(APIVersion(2, 28)))
def test_check_safe_for_pipette_movement_partial_tip_version_gate_above(
    decoy: Decoy,
    mock_state_view: StateView,
    api_version: APIVersion,
) -> None:
    """It should pass when location is a well and API version is at or over 2.28."""
    decoy.when(
        mock_state_view.pipettes.get_is_partially_configured("pipette-id")
    ).then_return(True)

    decoy.when(
        mock_state_view.geometry.get_checked_tip_drop_location(
            pipette_id="pipette-id",
            labware_id="destination-labware-id",
            well_location=DropTipWellLocation(),
            api_version_allows_partial_return_tip=True,
        )
    ).then_return(WellLocation(origin=WellOrigin.TOP, offset=WellOffset(z=10)))

    decoy.when(
        mock_state_view.geometry.get_well_position(
            labware_id="destination-labware-id",
            well_name="A2",
            well_location=WellLocation(origin=WellOrigin.TOP, offset=WellOffset(z=10)),
            pipette_id="pipette-id",
        )
    ).then_return(Point(x=123, y=123, z=123))

    decoy.when(mock_state_view.pipettes.get_primary_nozzle("pipette-id")).then_return(
        "A12"
    )
    decoy.when(
        mock_state_view.motion.get_critical_point_for_wells_in_labware(
            "destination-labware-id"
        )
    ).then_return(None)
    decoy.when(
        mock_state_view.pipettes.get_pipette_bounds_at_specified_move_to_position(
            pipette_id="pipette-id",
            destination_position=Point(x=123, y=123, z=123),
            critical_point=None,
        )
    ).then_return(
        (
            Point(x=50, y=150, z=60),
            Point(x=150, y=50, z=60),
            Point(x=150, y=150, z=60),
            Point(x=50, y=50, z=60),
        )
    )
    decoy.when(mock_state_view.geometry.absolute_deck_extents).then_return(
        _AbsoluteRobotExtents(
            front_left={
                MountType.LEFT: Point(13.5, -60.5, 0.0),
                MountType.RIGHT: Point(-40.5, -60.5, 0.0),
            },
            back_right={
                MountType.LEFT: Point(463.7, 433.3, 0.0),
                MountType.RIGHT: Point(517.7, 433.3),
            },
            deck_extents=Point(477.2, 493.8, 0.0),
            padding_rear=-181.21,
            padding_front=55.8,
            padding_left_side=31.88,
            padding_right_side=-80.32,
        )
    )
    decoy.when(
        mock_state_view.geometry.get_ancestor_slot_name("destination-labware-id")
    ).then_return(DeckSlotName.SLOT_C2)
    decoy.when(
        adjacent_slots_getters.get_surrounding_slots(5, robot_type="OT-3 Standard")
    ).then_return(
        _MixedTypeSlots(
            regular_slots=[
                DeckSlotName.SLOT_D1,
                DeckSlotName.SLOT_D2,
                DeckSlotName.SLOT_C1,
                DeckSlotName.SLOT_B3,
            ],
            staging_slots=[StagingSlotName.SLOT_C4],
        )
    )
    decoy.when(
        mock_state_view.addressable_areas.get_addressable_area_position(
            addressable_area_name="C1", do_compatibility_check=False
        )
    ).then_return(Point(0, 100, 0))
    decoy.when(
        mock_state_view.addressable_areas.get_addressable_area_position(
            addressable_area_name="D1", do_compatibility_check=False
        )
    ).then_return(Point(0, 0, 0))
    decoy.when(
        mock_state_view.addressable_areas.get_addressable_area_position(
            addressable_area_name="D2", do_compatibility_check=False
        )
    ).then_return(Point(100, 0, 0))
    decoy.when(
        mock_state_view.addressable_areas.get_addressable_area_position(
            addressable_area_name="C4", do_compatibility_check=False
        )
    ).then_return(Point(200, 100, 0))
    decoy.when(
        mock_state_view.addressable_areas.get_addressable_area_bounding_box(
            addressable_area_name="C4", do_compatibility_check=False
        )
    ).then_return(Dimensions(90, 90, 0))
    decoy.when(
        mock_state_view.geometry.get_highest_z_in_slot(
            StagingSlotLocation(slotName=StagingSlotName.SLOT_C4)
        )
    ).then_return(50)
    decoy.when(
        mock_state_view.addressable_areas.get_addressable_area_position(
            addressable_area_name="B3", do_compatibility_check=False
        )
    ).then_return(Point(150, 200, 0))
    decoy.when(
        mock_state_view.addressable_areas.get_addressable_area_bounding_box(
            addressable_area_name="B3", do_compatibility_check=False
        )
    ).then_return(Dimensions(90, 90, 0))
    for slot_name in [DeckSlotName.SLOT_C1, DeckSlotName.SLOT_D1, DeckSlotName.SLOT_D2]:
        decoy.when(
            mock_state_view.geometry.get_highest_z_in_slot(
                DeckSlotLocation(slotName=slot_name)
            )
        ).then_return(50)
        decoy.when(
            mock_state_view.addressable_areas.get_addressable_area_bounding_box(
                addressable_area_name=slot_name.id, do_compatibility_check=False
            )
        ).then_return(Dimensions(90, 90, 0))

    pipette_movement_conflict.check_safe_for_pipette_movement(
        engine_state=mock_state_view,
        pipette_id="pipette-id",
        labware_id="destination-labware-id",
        well_name="A2",
        well_location=DropTipWellLocation(),
        version=api_version,
    )


@pytest.mark.parametrize(
    ("robot_type", "deck_type"),
    [
        ("OT-3 Standard", DeckType.OT3_STANDARD),
    ],
)
@pytest.mark.parametrize(
    "api_version", versions_below(APIVersion(2, 28), flex_only=True)
)
def test_check_safe_for_pipette_movement_partial_tip_version_gate_below(
    decoy: Decoy,
    mock_state_view: StateView,
    api_version: APIVersion,
) -> None:
    """It should pass when location is a well and API version is below 2.28."""
    decoy.when(
        mock_state_view.pipettes.get_is_partially_configured("pipette-id")
    ).then_return(True)

    decoy.when(
        mock_state_view.geometry.get_checked_tip_drop_location(
            pipette_id="pipette-id",
            labware_id="destination-labware-id",
            well_location=DropTipWellLocation(),
            api_version_allows_partial_return_tip=False,
        )
    ).then_return(WellLocation(origin=WellOrigin.TOP, offset=WellOffset(z=10)))

    decoy.when(
        mock_state_view.geometry.get_well_position(
            labware_id="destination-labware-id",
            well_name="A2",
            well_location=WellLocation(origin=WellOrigin.TOP, offset=WellOffset(z=10)),
            pipette_id="pipette-id",
        )
    ).then_return(Point(x=123, y=123, z=123))

    decoy.when(mock_state_view.pipettes.get_primary_nozzle("pipette-id")).then_return(
        "A12"
    )
    decoy.when(
        mock_state_view.motion.get_critical_point_for_wells_in_labware(
            "destination-labware-id"
        )
    ).then_return(None)
    decoy.when(
        mock_state_view.pipettes.get_pipette_bounds_at_specified_move_to_position(
            pipette_id="pipette-id",
            destination_position=Point(x=123, y=123, z=123),
            critical_point=None,
        )
    ).then_return(
        (
            Point(x=50, y=150, z=60),
            Point(x=150, y=50, z=60),
            Point(x=150, y=150, z=60),
            Point(x=50, y=50, z=60),
        )
    )
    decoy.when(mock_state_view.geometry.absolute_deck_extents).then_return(
        _AbsoluteRobotExtents(
            front_left={
                MountType.LEFT: Point(13.5, -60.5, 0.0),
                MountType.RIGHT: Point(-40.5, -60.5, 0.0),
            },
            back_right={
                MountType.LEFT: Point(463.7, 433.3, 0.0),
                MountType.RIGHT: Point(517.7, 433.3),
            },
            deck_extents=Point(477.2, 493.8, 0.0),
            padding_rear=-181.21,
            padding_front=55.8,
            padding_left_side=31.88,
            padding_right_side=-80.32,
        )
    )
    decoy.when(
        mock_state_view.geometry.get_ancestor_slot_name("destination-labware-id")
    ).then_return(DeckSlotName.SLOT_C2)
    decoy.when(
        adjacent_slots_getters.get_surrounding_slots(5, robot_type="OT-3 Standard")
    ).then_return(
        _MixedTypeSlots(
            regular_slots=[
                DeckSlotName.SLOT_D1,
                DeckSlotName.SLOT_D2,
                DeckSlotName.SLOT_C1,
                DeckSlotName.SLOT_B3,
            ],
            staging_slots=[StagingSlotName.SLOT_C4],
        )
    )
    decoy.when(
        mock_state_view.addressable_areas.get_addressable_area_position(
            addressable_area_name="C1", do_compatibility_check=False
        )
    ).then_return(Point(0, 100, 0))
    decoy.when(
        mock_state_view.addressable_areas.get_addressable_area_position(
            addressable_area_name="D1", do_compatibility_check=False
        )
    ).then_return(Point(0, 0, 0))
    decoy.when(
        mock_state_view.addressable_areas.get_addressable_area_position(
            addressable_area_name="D2", do_compatibility_check=False
        )
    ).then_return(Point(100, 0, 0))
    decoy.when(
        mock_state_view.addressable_areas.get_addressable_area_position(
            addressable_area_name="C4", do_compatibility_check=False
        )
    ).then_return(Point(200, 100, 0))
    decoy.when(
        mock_state_view.addressable_areas.get_addressable_area_bounding_box(
            addressable_area_name="C4", do_compatibility_check=False
        )
    ).then_return(Dimensions(90, 90, 0))
    decoy.when(
        mock_state_view.geometry.get_highest_z_in_slot(
            StagingSlotLocation(slotName=StagingSlotName.SLOT_C4)
        )
    ).then_return(50)
    decoy.when(
        mock_state_view.addressable_areas.get_addressable_area_position(
            addressable_area_name="B3", do_compatibility_check=False
        )
    ).then_return(Point(150, 200, 0))
    decoy.when(
        mock_state_view.addressable_areas.get_addressable_area_bounding_box(
            addressable_area_name="B3", do_compatibility_check=False
        )
    ).then_return(Dimensions(90, 90, 0))
    for slot_name in [DeckSlotName.SLOT_C1, DeckSlotName.SLOT_D1, DeckSlotName.SLOT_D2]:
        decoy.when(
            mock_state_view.geometry.get_highest_z_in_slot(
                DeckSlotLocation(slotName=slot_name)
            )
        ).then_return(50)
        decoy.when(
            mock_state_view.addressable_areas.get_addressable_area_bounding_box(
                addressable_area_name=slot_name.id, do_compatibility_check=False
            )
        ).then_return(Dimensions(90, 90, 0))

    pipette_movement_conflict.check_safe_for_pipette_movement(
        engine_state=mock_state_view,
        pipette_id="pipette-id",
        labware_id="destination-labware-id",
        well_name="A2",
        well_location=DropTipWellLocation(),
        version=api_version,
    )



"""Command models to retrieve a labware from a Flex Stacker."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Literal, Union

from pydantic import BaseModel, Field
from pydantic.json_schema import SkipJsonSchema
from typing_extensions import Type

from opentrons_shared_data.errors.exceptions import (
    FlexStackerHopperLabwareError,
    FlexStackerShuttleLabwareError,
    FlexStackerShuttleMissingError,
    FlexStackerShuttleNotEmptyError,
    FlexStackerStallError,
)

from ...errors import (
    CannotPerformModuleAction,
    ErrorOccurrence,
    FlexStackerLabwarePoolNotYetDefinedError,
    LocationIsOccupiedError,
)
from ...resources import ModelUtils
from ...state import update_types
from ...types import (
    InStackerHopperLocation,
    LabwareLocationSequence,
    ModuleLocation,
)
from ..command import (
    AbstractCommandImpl,
    BaseCommand,
    BaseCommandCreate,
    DefinedErrorData,
    SuccessData,
)
from .common import (
    FlexStackerHopperError,
    FlexStackerLabwareRetrieveError,
    FlexStackerShuttleError,
    FlexStackerShuttleOccupiedError,
    FlexStackerStallOrCollisionError,
    adapter_location_sequence,
    build_retrieve_labware_move_updates,
    labware_locations_for_group,
    lid_location_sequence,
    primary_location_sequence,
)

if TYPE_CHECKING:
    from opentrons.protocol_engine.execution import EquipmentHandler
    from opentrons.protocol_engine.state.state import StateView

RetrieveCommandType = Literal["flexStacker/retrieve"]

RecoverableExceptions = Union[
    FlexStackerStallError,
    FlexStackerShuttleMissingError,
    FlexStackerHopperLabwareError,
    FlexStackerShuttleLabwareError,
    FlexStackerShuttleNotEmptyError,
]


def _remove_default(s: dict[str, Any]) -> None:
    s.pop("default", None)


class RetrieveParams(BaseModel):
    """Input parameters for a labware retrieval command."""

    moduleId: str = Field(
        ...,
        description="Unique ID of the Flex Stacker.",
    )
    labwareId: str | SkipJsonSchema[None] = Field(
        None,
        description="Do not use. Present for internal backward compatibility.",
        json_schema_extra=_remove_default,
    )
    displayName: str | SkipJsonSchema[None] = Field(
        None,
        description="Do not use. Present for internal backward compatibility.",
        json_schema_extra=_remove_default,
    )
    adapterId: str | SkipJsonSchema[None] = Field(
        None,
        description="Do not use. Present for internal backward compatibility.",
        json_schema_extra=_remove_default,
    )
    lidId: str | SkipJsonSchema[None] = Field(
        None,
        description="Do not use. Present for internal backward compatibility.",
        json_schema_extra=_remove_default,
    )


class RetrieveResult(BaseModel):
    """Result data from a labware retrieval command."""

    labwareId: str = Field(
        ...,
        description="The labware ID of the primary retrieved labware.",
    )
    adapterId: str | SkipJsonSchema[None] = Field(
        None,
        description="The optional Adapter Labware ID of the adapter under a primary labware.",
    )
    lidId: str | SkipJsonSchema[None] = Field(
        None,
        description="The optional Lid Labware ID of the lid on a primary labware.",
    )
    primaryLocationSequence: LabwareLocationSequence = Field(
        ..., description="The new location of the just-retrieved."
    )
    lidLocationSequence: LabwareLocationSequence | SkipJsonSchema[None] = Field(
        None,
        description="The new location of the just-retrieved adapter labware under a primary labware.",
    )
    adapterLocationSequence: LabwareLocationSequence | SkipJsonSchema[None] = Field(
        None,
        description="The new location of the just-retrieved lid labware on a primary labware.",
    )
    originalPrimaryLocationSequence: LabwareLocationSequence = Field(
        ..., description="The original location of the just-retrieved primary labware"
    )
    originalAdapterLocationSequence: LabwareLocationSequence | SkipJsonSchema[None] = (
        Field(None, description="The original location of an adapter labware if any")
    )
    originalLidLocationSequence: LabwareLocationSequence | SkipJsonSchema[None] = Field(
        None, description="The original location of a lid labware if any"
    )
    primaryLabwareURI: str = Field(
        ...,
        description="The labware definition URI of the primary labware.",
    )
    adapterLabwareURI: str | SkipJsonSchema[None] = Field(
        None,
        description="The labware definition URI of the adapter labware.",
    )
    lidLabwareURI: str | SkipJsonSchema[None] = Field(
        None,
        description="The labware definition URI of the lid labware.",
    )


_ExecuteReturn = Union[
    SuccessData[RetrieveResult],
    DefinedErrorData[FlexStackerStallOrCollisionError]
    | DefinedErrorData[FlexStackerShuttleError]
    | DefinedErrorData[FlexStackerHopperError]
    | DefinedErrorData[FlexStackerLabwareRetrieveError]
    | DefinedErrorData[FlexStackerShuttleOccupiedError],
]


class RetrieveImpl(AbstractCommandImpl[RetrieveParams, _ExecuteReturn]):
    """Implementation of a labware retrieval command."""

    def __init__(
        self,
        state_view: StateView,
        equipment: EquipmentHandler,
        model_utils: ModelUtils,
        **kwargs: object,
    ) -> None:
        self._state_view = state_view
        self._equipment = equipment
        self._model_utils = model_utils

    def handle_recoverable_error(
        self,
        error: RecoverableExceptions,
        intended_id: str,
        state_update: update_types.StateUpdate,
    ) -> (
        DefinedErrorData[FlexStackerStallOrCollisionError]
        | DefinedErrorData[FlexStackerShuttleError]
        | DefinedErrorData[FlexStackerHopperError]
        | DefinedErrorData[FlexStackerLabwareRetrieveError]
        | DefinedErrorData[FlexStackerShuttleOccupiedError]
    ):
        """Handle a recoverable error raised during command execution."""
        error_map = {
            FlexStackerStallError: FlexStackerStallOrCollisionError,
            FlexStackerShuttleMissingError: FlexStackerShuttleError,
            FlexStackerHopperLabwareError: FlexStackerHopperError,
            FlexStackerShuttleLabwareError: FlexStackerLabwareRetrieveError,
            FlexStackerShuttleNotEmptyError: FlexStackerShuttleOccupiedError,
        }
        return DefinedErrorData(
            public=error_map[type(error)](
                id=self._model_utils.generate_id(),
                createdAt=self._model_utils.get_timestamp(),
                wrappedErrors=[
                    ErrorOccurrence.from_failed(
                        id=self._model_utils.generate_id(),
                        createdAt=self._model_utils.get_timestamp(),
                        error=error,
                    )
                ],
                errorInfo={"labwareId": intended_id},
            ),
            state_update_if_false_positive=state_update,
        )

    async def execute(self, params: RetrieveParams) -> _ExecuteReturn:
        """Execute the labware retrieval command."""
        stacker_state = self._state_view.modules.get_flex_stacker_substate(
            params.moduleId
        )
        location = self._state_view.modules.get_location(params.moduleId)

        if stacker_state.pool_primary_definition is None:
            raise FlexStackerLabwarePoolNotYetDefinedError(
                f"The Flex Stacker in {location} must be configured before labware can be retrieved."
            )

        if len(stacker_state.contained_labware_bottom_first) == 0:
            raise CannotPerformModuleAction(
                message=f"Cannot retrieve labware from Flex Stacker in {location} because it contains no labware"
            )

        stacker_loc = ModuleLocation(moduleId=params.moduleId)

        try:
            self._state_view.labware.raise_if_labware_in_location(stacker_loc)
        except LocationIsOccupiedError:
            raise CannotPerformModuleAction(
                f"Cannot retrieve a labware from Flex Stacker in {location} if the carriage is occupied"
            )

        to_retrieve = stacker_state.contained_labware_bottom_first[0]
        remaining = stacker_state.contained_labware_bottom_first[1:]

        locations, offsets = build_retrieve_labware_move_updates(
            to_retrieve, stacker_state, self._state_view
        )

        # Update the state to reflect the labware is now in the Flex Stacker slot
        # todo(chb, 2025-02-19): This ModuleLocation piece should probably instead be an AddressableAreaLocation
        # but that has implications for where labware are set by things like module.load_labware(..) and what
        # happens when we move labware.
        stacker_area = (
            self._state_view.modules.ensure_and_convert_module_fixture_location(
                deck_slot=self._state_view.modules.get_location(
                    params.moduleId
                ).slotName,
                model=self._state_view.modules.get(params.moduleId).model,
            )
        )
        original_locations = labware_locations_for_group(
            to_retrieve, [InStackerHopperLocation(moduleId=params.moduleId)]
        )
        new_locations = labware_locations_for_group(
            to_retrieve,
            self._state_view.geometry.get_predicted_location_sequence(
                ModuleLocation(moduleId=params.moduleId)
            ),
        )

        state_update = (
            update_types.StateUpdate()
            .set_batch_labware_location(
                new_locations_by_id=locations, new_offset_ids_by_id=offsets
            )
            .update_flex_stacker_contained_labware(params.moduleId, remaining)
            .set_addressable_area_used(stacker_area)
        )

        # Allow propagation of ModuleNotAttachedError.
        stacker_hw = self._equipment.get_module_hardware_api(stacker_state.module_id)
        if stacker_hw is not None:
            try:
                stacker_hw.set_stacker_identify(True)
                await stacker_hw.dispense_labware(
                    labware_height=stacker_state.get_pool_height_minus_overlap()
                )
            except (
                FlexStackerStallError,
                FlexStackerShuttleMissingError,
                FlexStackerHopperLabwareError,
                FlexStackerShuttleLabwareError,
                FlexStackerShuttleNotEmptyError,
            ) as e:
                return self.handle_recoverable_error(
                    e, to_retrieve.primaryLabwareId, state_update
                )

        if stacker_hw is not None:
            stacker_hw.set_stacker_identify(False)

        return SuccessData(
            public=RetrieveResult.model_construct(
                labwareId=to_retrieve.primaryLabwareId,
                adapterId=to_retrieve.adapterLabwareId,
                lidId=to_retrieve.lidLabwareId,
                primaryLocationSequence=primary_location_sequence(new_locations),
                adapterLocationSequence=adapter_location_sequence(new_locations),
                lidLocationSequence=lid_location_sequence(new_locations),
                originalPrimaryLocationSequence=primary_location_sequence(
                    original_locations
                ),
                originalAdapterLocationSequence=adapter_location_sequence(
                    original_locations
                ),
                originalLidLocationSequence=lid_location_sequence(original_locations),
                primaryLabwareURI=self._state_view.labware.get_uri_from_definition(
                    stacker_state.pool_primary_definition
                ),
                adapterLabwareURI=self._state_view.labware.get_uri_from_definition_unless_none(
                    stacker_state.pool_adapter_definition
                ),
                lidLabwareURI=self._state_view.labware.get_uri_from_definition_unless_none(
                    stacker_state.pool_lid_definition
                ),
            ),
            state_update=state_update,
        )


class Retrieve(BaseCommand[RetrieveParams, RetrieveResult, ErrorOccurrence]):
    """A command to retrieve a labware from a Flex Stacker."""

    commandType: RetrieveCommandType = "flexStacker/retrieve"
    params: RetrieveParams
    result: RetrieveResult | None = None

    _ImplementationCls: Type[RetrieveImpl] = RetrieveImpl


class RetrieveCreate(BaseCommandCreate[RetrieveParams]):
    """A request to execute a Flex Stacker retrieve command."""

    commandType: RetrieveCommandType = "flexStacker/retrieve"
    params: RetrieveParams

    _CommandCls: Type[Retrieve] = Retrieve



"""Tests for well math."""

import json
import pathlib
from collections import OrderedDict
from itertools import chain
from typing import Any, cast

import pytest

from opentrons_shared_data.pipette.pipette_definition import ValidNozzleMaps

from .. import pipette_fixtures
from opentrons.hardware_control.nozzle_manager import NozzleMap
from opentrons.protocol_engine.state._well_math import (
    nozzles_per_well,
    wells_covered_dense,
    wells_covered_sparse,
)
from opentrons.types import Point

_96_FULL_MAP = NozzleMap.build(
    physical_nozzles=pipette_fixtures.NINETY_SIX_MAP,
    physical_rows=pipette_fixtures.NINETY_SIX_ROWS,
    physical_columns=pipette_fixtures.NINETY_SIX_COLS,
    starting_nozzle="A1",
    back_left_nozzle="A1",
    front_right_nozzle="H12",
    valid_nozzle_maps=ValidNozzleMaps(
        maps={
            "Full": sum(
                [
                    pipette_fixtures.NINETY_SIX_ROWS["A"],
                    pipette_fixtures.NINETY_SIX_ROWS["B"],
                    pipette_fixtures.NINETY_SIX_ROWS["C"],
                    pipette_fixtures.NINETY_SIX_ROWS["D"],
                    pipette_fixtures.NINETY_SIX_ROWS["E"],
                    pipette_fixtures.NINETY_SIX_ROWS["F"],
                    pipette_fixtures.NINETY_SIX_ROWS["G"],
                    pipette_fixtures.NINETY_SIX_ROWS["H"],
                ],
                [],
            )
        }
    ),
)
_96_COL1_MAP = NozzleMap.build(
    physical_nozzles=pipette_fixtures.NINETY_SIX_MAP,
    physical_rows=pipette_fixtures.NINETY_SIX_ROWS,
    physical_columns=pipette_fixtures.NINETY_SIX_COLS,
    starting_nozzle="A1",
    back_left_nozzle="A1",
    front_right_nozzle="H1",
    valid_nozzle_maps=ValidNozzleMaps(
        maps={"Column1": pipette_fixtures.NINETY_SIX_COLS["1"]}
    ),
)

_96_COL12_MAP = NozzleMap.build(
    physical_nozzles=pipette_fixtures.NINETY_SIX_MAP,
    physical_rows=pipette_fixtures.NINETY_SIX_ROWS,
    physical_columns=pipette_fixtures.NINETY_SIX_COLS,
    starting_nozzle="A12",
    back_left_nozzle="A12",
    front_right_nozzle="H12",
    valid_nozzle_maps=ValidNozzleMaps(
        maps={"Column12": pipette_fixtures.NINETY_SIX_COLS["12"]}
    ),
)

_96_SINGLE_FR_MAP = NozzleMap.build(
    physical_nozzles=pipette_fixtures.NINETY_SIX_MAP,
    physical_rows=pipette_fixtures.NINETY_SIX_ROWS,
    physical_columns=pipette_fixtures.NINETY_SIX_COLS,
    starting_nozzle="H12",
    back_left_nozzle="H12",
    front_right_nozzle="H12",
    valid_nozzle_maps=ValidNozzleMaps(maps={"Single": ["H12"]}),
)
_96_SINGLE_BL_MAP = NozzleMap.build(
    physical_nozzles=pipette_fixtures.NINETY_SIX_MAP,
    physical_rows=pipette_fixtures.NINETY_SIX_ROWS,
    physical_columns=pipette_fixtures.NINETY_SIX_COLS,
    starting_nozzle="A1",
    back_left_nozzle="A1",
    front_right_nozzle="A1",
    valid_nozzle_maps=ValidNozzleMaps(maps={"Single": ["A1"]}),
)
_96_RECTANGLE_MAP = NozzleMap.build(
    physical_nozzles=pipette_fixtures.NINETY_SIX_MAP,
    physical_rows=pipette_fixtures.NINETY_SIX_ROWS,
    physical_columns=pipette_fixtures.NINETY_SIX_COLS,
    starting_nozzle="A1",
    back_left_nozzle="A1",
    front_right_nozzle="E2",
    valid_nozzle_maps=ValidNozzleMaps(
        maps={
            "Subrect": [
                "A1",
                "A2",
                "B1",
                "B2",
                "C1",
                "C2",
                "D1",
                "D2",
                "E1",
                "E2",
            ]
        }
    ),
)
_8_FULL_MAP = NozzleMap.build(
    physical_nozzles=pipette_fixtures.EIGHT_CHANNEL_MAP,
    physical_rows=pipette_fixtures.EIGHT_CHANNEL_ROWS,
    physical_columns=pipette_fixtures.EIGHT_CHANNEL_COLS,
    starting_nozzle="A1",
    back_left_nozzle="A1",
    front_right_nozzle="H1",
    valid_nozzle_maps=ValidNozzleMaps(
        maps={"Full": pipette_fixtures.EIGHT_CHANNEL_COLS["1"]}
    ),
)
_8_SINGLE_MAP = NozzleMap.build(
    physical_nozzles=pipette_fixtures.EIGHT_CHANNEL_MAP,
    physical_rows=pipette_fixtures.EIGHT_CHANNEL_ROWS,
    physical_columns=pipette_fixtures.EIGHT_CHANNEL_COLS,
    starting_nozzle="A1",
    back_left_nozzle="A1",
    front_right_nozzle="A1",
    valid_nozzle_maps=ValidNozzleMaps(maps={"Full": ["A1"]}),
)
_8_HALF_MAP = NozzleMap.build(
    physical_nozzles=pipette_fixtures.EIGHT_CHANNEL_MAP,
    physical_rows=pipette_fixtures.EIGHT_CHANNEL_ROWS,
    physical_columns=pipette_fixtures.EIGHT_CHANNEL_COLS,
    starting_nozzle="A1",
    back_left_nozzle="A1",
    front_right_nozzle="D1",
    valid_nozzle_maps=ValidNozzleMaps(maps={"Half": ["A1", "B1", "C1", "D1"]}),
)
_SINGLE_MAP = NozzleMap.build(
    physical_nozzles=OrderedDict((("A1", Point(0.0, 1.0, 2.0)),)),
    physical_rows=OrderedDict((("1", ["A1"]),)),
    physical_columns=OrderedDict((("A", ["A1"]),)),
    starting_nozzle="A1",
    back_left_nozzle="A1",
    front_right_nozzle="A1",
    valid_nozzle_maps=ValidNozzleMaps(maps={"Single": ["A1"]}),
)


def _fixture(fixture_name: str) -> Any:
    return json.load(
        open(
            pathlib.Path(__file__).parent
            / f"../../../../../shared-data/labware/fixtures/2/{fixture_name}.json"
        )
    )


@pytest.fixture
def _96_wells() -> list[list[str]]:
    return fixture_map("fixture_96_plate")


@pytest.fixture
def _384_wells() -> list[list[str]]:
    return fixture_map("fixture_384_plate")


@pytest.fixture
def _12_reservoir() -> list[list[str]]:
    return fixture_map("fixture_12_trough_v2")


@pytest.fixture
def _1_reservoir() -> list[list[str]]:
    return [["A1"]]


def all_wells(fixture_name: str) -> list[str]:
    """All wells in a labware as a flat list."""
    ordering = fixture_map(fixture_name)
    return list(chain(*ordering))


def fixture_map(fixture_name: str) -> list[list[str]]:
    """The ordering map."""
    return cast(list[list[str]], _fixture(fixture_name)["ordering"])


@pytest.mark.parametrize(
    "nozzle_map,target_well,result",
    [
        # these configurations have all the nozzles on/in the wellplate
        (_SINGLE_MAP, "A1", ["A1"]),
        (_SINGLE_MAP, "D7", ["D7"]),
        (_8_FULL_MAP, "A1", ["A1", "B1", "C1", "D1", "E1", "F1", "G1", "H1"]),
        (_8_FULL_MAP, "A10", ["A10", "B10", "C10", "D10", "E10", "F10", "G10", "H10"]),
        (_96_FULL_MAP, "A1", all_wells("fixture_96_plate")),
        (
            _96_RECTANGLE_MAP,
            "C8",
            ["C8", "D8", "E8", "F8", "G8", "C9", "D9", "E9", "F9", "G9"],
        ),
        # these configurations have some nozzles below or to the right
        (_8_FULL_MAP, "D1", ["D1", "E1", "F1", "G1", "H1"]),
        (
            _96_FULL_MAP,
            "C8",
            [
                well
                for well in all_wells("fixture_96_plate")
                if ord(well[0]) >= ord("C") and int(well[1:]) >= 8
            ],
        ),
    ],
)
def test_wells_covered_dense_96(
    nozzle_map: NozzleMap,
    target_well: str,
    result: list[str],
    _96_wells: list[list[str]],
) -> None:
    """It should calculate well coverage for an SBS 96."""
    assert (
        list(
            wells_covered_dense(
                nozzle_map.columns,
                nozzle_map.rows,
                nozzle_map.starting_nozzle,
                target_well,
                _96_wells,
            )
        )
        == result
    )


@pytest.mark.parametrize(
    "nozzle_map,target_well,result",
    [
        # these configurations have all the nozzles on/in the wellplate
        (_SINGLE_MAP, "A1", ["A1"]),
        (_SINGLE_MAP, "D7", ["D7"]),
        (_8_FULL_MAP, "A1", ["A1", "C1", "E1", "G1", "I1", "K1", "M1", "O1"]),
        (_8_FULL_MAP, "B1", ["B1", "D1", "F1", "H1", "J1", "L1", "N1", "P1"]),
        (_8_FULL_MAP, "A10", ["A10", "C10", "E10", "G10", "I10", "K10", "M10", "O10"]),
        # well offsets inside the four-well clusters that are the size of each 96-well well
        (
            _96_FULL_MAP,
            "A1",
            [
                well
                for well in all_wells("fixture_384_plate")
                if well[0] in "ACEGIKMO"
                and int(well[1:]) in [1, 3, 5, 7, 9, 11, 13, 15, 17, 19, 21, 23]
            ],
        ),
        (
            _96_FULL_MAP,
            "A2",
            [
                well
                for well in all_wells("fixture_384_plate")
                if well[0] in "ACEGIKMO"
                and int(well[1:]) in [2, 4, 6, 8, 10, 12, 14, 16, 18, 20, 22, 24]
            ],
        ),
        (
            _96_FULL_MAP,
            "B1",
            [
                well
                for well in all_wells("fixture_384_plate")
                if well[0] in "BDFHJLNP"
                and int(well[1:]) in [1, 3, 5, 7, 9, 11, 13, 15, 17, 19, 21, 23]
            ],
        ),
        (
            _96_FULL_MAP,
            "B2",
            [
                well
                for well in all_wells("fixture_384_plate")
                if well[0] in "BDFHJLNP"
                and int(well[1:]) in [2, 4, 6, 8, 10, 12, 14, 16, 18, 20, 22, 24]
            ],
        ),
        (
            _96_RECTANGLE_MAP,
            "C8",
            ["C8", "E8", "G8", "I8", "K8", "C10", "E10", "G10", "I10", "K10"],
        ),
        (
            _96_RECTANGLE_MAP,
            "C9",
            ["C9", "E9", "G9", "I9", "K9", "C11", "E11", "G11", "I11", "K11"],
        ),
        (
            _96_RECTANGLE_MAP,
            "D8",
            ["D8", "F8", "H8", "J8", "L8", "D10", "F10", "H10", "J10", "L10"],
        ),
        (
            _96_RECTANGLE_MAP,
            "D9",
            ["D9", "F9", "H9", "J9", "L9", "D11", "F11", "H11", "J11", "L11"],
        ),
    ],
)
def test_wells_covered_dense_384(
    nozzle_map: NozzleMap,
    target_well: str,
    result: list[str],
    _384_wells: list[list[str]],
) -> None:
    """It should calculate well coverage for an SBS 384."""
    assert (
        list(
            wells_covered_dense(
                nozzle_map.columns,
                nozzle_map.rows,
                nozzle_map.starting_nozzle,
                target_well,
                _384_wells,
            )
        )
        == result
    )


@pytest.mark.parametrize(
    "nozzle_map,target_well,result",
    [
        (_SINGLE_MAP, "A1", ["A1"]),
        (_SINGLE_MAP, "A8", ["A8"]),
        (_8_FULL_MAP, "A1", ["A1"]),
        (_8_FULL_MAP, "A8", ["A8"]),
        (
            _96_FULL_MAP,
            "A1",
            all_wells("fixture_12_trough_v2"),
        ),
        (
            _96_FULL_MAP,
            "A8",
            [well for well in all_wells("fixture_12_trough_v2") if int(well[1:]) >= 8],
        ),
        (_96_RECTANGLE_MAP, "A1", ["A1", "A2"]),
        (_96_RECTANGLE_MAP, "A8", ["A8", "A9"]),
    ],
)
def test_wells_covered_sparse_12(
    nozzle_map: NozzleMap,
    target_well: str,
    result: list[str],
    _12_reservoir: list[list[str]],
) -> None:
    """It should calculate well coverage for a 12 column reservoir."""
    assert (
        list(
            wells_covered_sparse(
                nozzle_map.columns,
                nozzle_map.rows,
                nozzle_map.starting_nozzle,
                target_well,
                _12_reservoir,
            )
        )
        == result
    )


@pytest.mark.parametrize(
    "nozzle_map",
    [
        _SINGLE_MAP,
        _8_FULL_MAP,
        _96_FULL_MAP,
        _96_RECTANGLE_MAP,
    ],
)
def test_wells_covered_sparse_1(
    nozzle_map: NozzleMap, _1_reservoir: list[list[str]]
) -> None:
    """It should calculate well coverage for a 1 column reservoir."""
    assert list(
        wells_covered_sparse(
            nozzle_map.columns,
            nozzle_map.rows,
            nozzle_map.starting_nozzle,
            "A1",
            _1_reservoir,
        )
    ) == ["A1"]


@pytest.mark.parametrize(
    "nozzle_map",
    [
        _SINGLE_MAP,
        _8_FULL_MAP,
        _96_FULL_MAP,
        _96_RECTANGLE_MAP,
        _8_SINGLE_MAP,
        _96_SINGLE_BL_MAP,
        _96_SINGLE_FR_MAP,
    ],
)
@pytest.mark.parametrize("fixture_name", ["fixture_384_plate", "fixture_96_plate"])
def test_nozzles_per_well_dense_force_1(
    nozzle_map: NozzleMap, fixture_name: str
) -> None:
    """It should calculate nozzles per well for SBS dense labware."""
    all_fixture_wells = all_wells(fixture_name)
    # it's a bit unreasonable to test every well of a 384 plate so walk the diagonal
    well_name = "A1"
    while True:
        if well_name not in all_fixture_wells:
            break
        assert nozzles_per_well(nozzle_map, well_name, fixture_map(fixture_name)) == 1
        well_name = f"{chr(ord(well_name[0]) + 1)}{str(int(well_name[1:]) + 1)}"


@pytest.mark.parametrize(
    "nozzle_map,target_well,result",
    [
        (_SINGLE_MAP, "A1", 1),
        (_SINGLE_MAP, "A12", 1),
        (_8_FULL_MAP, "A1", 8),
        (_96_FULL_MAP, "A1", 8),
        (_96_FULL_MAP, "A12", 8),
        (_96_RECTANGLE_MAP, "A4", 5),
    ],
)
def test_nozzles_per_well_12column(
    nozzle_map: NozzleMap, target_well: str, result: int
) -> None:
    """It should calculate nozzles per well for a 12 column."""
    assert (
        nozzles_per_well(nozzle_map, target_well, fixture_map("fixture_12_trough_v2"))
        == result
    )


@pytest.mark.parametrize(
    "nozzle_map,result",
    [
        (_SINGLE_MAP, 1),
        (_SINGLE_MAP, 1),
        (_8_FULL_MAP, 8),
        (_96_FULL_MAP, 96),
        (_96_FULL_MAP, 96),
        (_96_RECTANGLE_MAP, 10),
    ],
)
def test_nozzles_per_well_1column(nozzle_map: NozzleMap, result: int) -> None:
    """It should calculate nozzles per well for a 1-well reservoir."""
    assert nozzles_per_well(nozzle_map, "A1", [["A1"]]) == result



"""Tests for modules."""

import typing

import pytest

from opentrons import protocol_api


@pytest.mark.parametrize(
    "simulated_protocol_context", [("2.21", "Flex")], indirect=True
)
def test_absorbance_reader_labware_load_conflict(
    simulated_protocol_context: protocol_api.ProtocolContext,
) -> None:
    """It should prevent loading a labware onto a closed absorbance reader."""
    module = simulated_protocol_context.load_module("absorbanceReaderV1", "A3")

    # The lid should be treated as initially closed.
    with pytest.raises(Exception):
        module.load_labware("opentrons_96_wellplate_200ul_pcr_full_skirt")

    module.open_lid()  # type: ignore[union-attr]
    # Should not raise after opening the lid.
    labware_1 = module.load_labware("opentrons_96_wellplate_200ul_pcr_full_skirt")

    simulated_protocol_context.move_labware(labware_1, protocol_api.OFF_DECK)

    # Should raise after closing the lid again.
    module.close_lid()  # type: ignore[union-attr]
    with pytest.raises(Exception):
        module.load_labware("opentrons_96_wellplate_200ul_pcr_full_skirt")


@pytest.mark.parametrize(
    "simulated_protocol_context", [("2.21", "Flex")], indirect=True
)
def test_absorbance_reader_labware_move_conflict(
    simulated_protocol_context: protocol_api.ProtocolContext,
) -> None:
    """It should prevent moving a labware onto a closed absorbance reader."""
    module = simulated_protocol_context.load_module("absorbanceReaderV1", "A3")
    labware = simulated_protocol_context.load_labware(
        "opentrons_96_wellplate_200ul_pcr_full_skirt", "A1"
    )

    with pytest.raises(Exception):
        # The lid should be treated as initially closed.
        simulated_protocol_context.move_labware(labware, module, use_gripper=True)

    module.open_lid()  # type: ignore[union-attr]
    # Should not raise after opening the lid.
    simulated_protocol_context.move_labware(labware, module, use_gripper=True)

    simulated_protocol_context.move_labware(labware, "A1", use_gripper=True)

    # Should raise after closing the lid again.
    module.close_lid()  # type: ignore[union-attr]
    with pytest.raises(Exception):
        simulated_protocol_context.move_labware(labware, module, use_gripper=True)


@pytest.mark.parametrize(
    "simulated_protocol_context", [("2.21", "Flex")], indirect=True
)
def test_absorbance_reader_read_preconditions(
    simulated_protocol_context: protocol_api.ProtocolContext,
) -> None:
    """Test the preconditions for triggering an absorbance reader read."""
    module = typing.cast(
        protocol_api.AbsorbanceReaderContext,
        simulated_protocol_context.load_module("absorbanceReaderV1", "A3"),
    )

    with pytest.raises(Exception, match="initialize"):
        module.read()  # .initialize() must be called first.

    with pytest.raises(Exception, match="close"):
        module.initialize("single", [500])  # .close_lid() must be called first.

    module.close_lid()
    module.initialize("single", [500])

    module.read()  # Should not raise now.



"""Tests for Protocol API engine parameter validation."""

from typing import List, Optional

import pytest

from tests.opentrons.protocol_api import versions_between

from opentrons.protocol_api.core.engine import load_labware_params as subject
from opentrons.protocol_engine.state.labware import LabwareLoadParams
from opentrons.protocols.api_support.constants import OPENTRONS_NAMESPACE
from opentrons.protocols.api_support.types import APIVersion


@pytest.mark.parametrize(
    argnames=[
        "load_name",
        "namespace",
        "version",
        "custom_labware_params",
        "expected_namespace",
        "expected_version",
        "current_api_version",
    ],
    argvalues=[
        ("hello", "world", 123, [], "world", 123, APIVersion(2, 15)),
        ("hello", "world", None, [], "world", 1, APIVersion(2, 22)),
        ("hello", None, 123, [], OPENTRONS_NAMESPACE, 123, APIVersion(2, 14)),
        ("hello", None, None, [], OPENTRONS_NAMESPACE, 1, APIVersion(2, 16)),
        (
            "hello",
            "world",
            None,
            [
                LabwareLoadParams("hello", "world", 123),
                LabwareLoadParams("hello", "foo", 123),
            ],
            "world",
            123,
            APIVersion(2, 17),
        ),
        (
            "hello",
            None,
            123,
            [
                LabwareLoadParams("hello", "world", 123),
                LabwareLoadParams("hello", "world", 456),
            ],
            "world",
            123,
            APIVersion(2, 19),
        ),
        (
            "hello",
            None,
            None,
            [
                LabwareLoadParams("hello", "world", 123),
                LabwareLoadParams("goodbye", "world", 123),
            ],
            "world",
            123,
            APIVersion(2, 21),
        ),
    ],
)
def test_resolve_load_labware_params(
    load_name: str,
    namespace: Optional[str],
    version: Optional[int],
    custom_labware_params: List[LabwareLoadParams],
    expected_namespace: str,
    expected_version: int,
    current_api_version: APIVersion,
) -> None:
    """It should get a namespace and version based on custom labware available or defaults."""
    result = subject.resolve(
        load_name, namespace, version, custom_labware_params, current_api_version
    )

    assert result == (expected_namespace, expected_version)


@pytest.mark.parametrize(
    ("api_version", "expected_labware_version"),
    [
        # Subject assumes api_version always >=2.14.
        *[
            (api_version, 2)
            for api_version in versions_between(
                low_inclusive_bound=APIVersion(2, 14),
                high_inclusive_bound=APIVersion(2, 22),
            )
        ],
        *[
            (api_version, 4)
            for api_version in versions_between(
                low_inclusive_bound=APIVersion(2, 25),
                high_exclusive_bound=(APIVersion(2, 26)),
            )
        ],
    ],
)
def test_default_labware_version_dependent_on_api_version(
    api_version: APIVersion, expected_labware_version: int
) -> None:
    """Test the default labware version when it's dependent on api_version.

    We only test this with a single "interesting" labware because otherwise we'd
    basically be rewriting the production code.
    """
    result = subject.resolve(
        load_name="corning_12_wellplate_6.9ml_flat",
        namespace="opentrons",
        version=None,
        custom_load_labware_params=[],
        api_version=api_version,
    )
    assert result == (OPENTRONS_NAMESPACE, expected_labware_version)


@pytest.mark.parametrize(
    "load_name",
    [
        # An arbitrary sampling of labware whose default versions are normally
        # dependent on api_version.
        "opentrons_24_aluminumblock_generic_2ml_screwcap",
        "opentrons_96_aluminumblock_generic_pcr_strip_200ul",
        "armadillo_96_wellplate_200ul_pcr_full_skirt",
        "corning_12_wellplate_6.9ml_flat",
        "corning_384_wellplate_112ul_flat",
        "nest_96_wellplate_100ul_pcr_full_skirt",
        "nest_96_wellplate_200ul_flat",
        "nest_96_wellplate_2ml_deep",
        "opentrons_96_wellplate_200ul_pcr_full_skirt",
        "biorad_96_wellplate_200ul_pcr",
        "ev_resin_tips_flex_96_labware",
        "ev_resin_tips_flex_96_tiprack_adapter",
    ],
)
@pytest.mark.parametrize("namespace", [OPENTRONS_NAMESPACE, None])
@pytest.mark.parametrize(
    ("version", "expected_version"),
    [
        (0, 0),
        (1, 1),
        (2, 2),
        (123456, 123456),
    ],
)
def test_explicit_version_wins_against_default(
    load_name: str,
    namespace: Optional[str],
    version: Optional[int],
    expected_version: int,
) -> None:
    """Test the returned labware version when an explicit version is given.

    If you pass an explicit version, it should use exactly that, even if there would
    normally be some other default version because of the given api_version.
    """
    result = subject.resolve(
        load_name=load_name,
        namespace=namespace,
        version=version,
        custom_load_labware_params=[],
        api_version=APIVersion(2, 23),
    )
    assert result == (OPENTRONS_NAMESPACE, expected_version)


def test_resolve_load_labware_params_raises() -> None:
    """It should raise if multiple custom labware params are provided."""
    with pytest.raises(subject.AmbiguousLoadLabwareParamsError):
        subject.resolve(
            load_name="hello",
            namespace="world",
            version=None,
            custom_load_labware_params=[
                LabwareLoadParams("hello", "world", 123),
                LabwareLoadParams("hello", "world", 456),
            ],
            api_version=APIVersion(2, 23),
        )



"""Tests for opentrons.protocol_reader.role_analyzer.RoleAnalyzer."""

from typing import List, Union

import pytest

from opentrons.protocol_reader.file_identifier import (
    IdentifiedData,
    IdentifiedFile,
    IdentifiedJsonMain,
    IdentifiedLabwareDefinition,
    IdentifiedPythonMain,
)
from opentrons.protocol_reader.file_reader_writer import BufferedFile
from opentrons.protocol_reader.role_analyzer import (
    RoleAnalysis,
    RoleAnalysisError,
    RoleAnalyzer,
)
from opentrons.protocols.api_support.types import APIVersion


def dummy_python_protocol_file(file_name: str) -> IdentifiedPythonMain:
    """Return a IdentifiedPythonMain with trivial placeholder data."""
    return IdentifiedPythonMain(
        original_file=BufferedFile(name=file_name, path=None, contents=b""),
        api_level=APIVersion(9001, 9001),
        robot_type="OT-2 Standard",
        metadata={},
    )


def dummy_json_protocol_file(file_name: str) -> IdentifiedJsonMain:
    """Return a IdentifiedJsonMain with trivial placeholder data."""
    return IdentifiedJsonMain(
        original_file=BufferedFile(
            name=file_name,
            contents=b"",
            path=None,
        ),
        unvalidated_json={},
        schema_version=123,
        robot_type="OT-2 Standard",
        metadata={},
    )


def dummy_labware_definition_file(file_name: str) -> IdentifiedLabwareDefinition:
    """Return a IdentifiedLabwareDefinition with trivial placeholder data."""
    return IdentifiedLabwareDefinition(
        original_file=BufferedFile(
            name=file_name,
            contents=b"",
            path=None,
        ),
        unvalidated_json={},
    )


def dummy_data_file(file_name: str) -> IdentifiedData:
    """Return an IdentifiedData with trivial placeholder data."""
    return IdentifiedData(
        original_file=BufferedFile(
            name=file_name,
            contents=b"",
            path=None,
        )
    )


def test_role_analysis_all_files() -> None:
    """It should return all member files."""
    main_file = dummy_python_protocol_file(file_name="main")
    labware_file_1 = dummy_labware_definition_file(file_name="labware_1")
    labware_file_2 = dummy_labware_definition_file(file_name="labware_2")
    data_file_1 = dummy_data_file(file_name="data_1")
    data_file_2 = dummy_data_file(file_name="data_2")
    subject = RoleAnalysis(
        main_file=main_file,
        labware_files=[labware_file_1, labware_file_2],
        data_files=[data_file_1, data_file_2],
    )
    assert subject.all_files == [
        main_file,
        labware_file_1,
        labware_file_2,
        data_file_1,
        data_file_2,
    ]


@pytest.mark.parametrize(
    "main_file",
    [
        dummy_python_protocol_file(file_name="main_file_name"),
        dummy_json_protocol_file(file_name="main_file_name"),
    ],
)
@pytest.mark.parametrize(
    "labware_files",
    [
        [],
        [
            IdentifiedLabwareDefinition(
                original_file=BufferedFile(
                    name="labware_1_file_name",
                    contents=b"",
                    path=None,
                ),
                unvalidated_json={},
            ),
            IdentifiedLabwareDefinition(
                original_file=BufferedFile(
                    name="labware_2_file_name",
                    contents=b"",
                    path=None,
                ),
                unvalidated_json={},
            ),
        ],
    ],
)
@pytest.mark.parametrize(
    "data_files",
    [
        [],
        [
            IdentifiedData(
                original_file=BufferedFile(
                    name="data_1_file_name",
                    contents=b"",
                    path=None,
                ),
            ),
            IdentifiedData(
                original_file=BufferedFile(
                    name="data_2_file_name",
                    contents=b"",
                    path=None,
                ),
            ),
        ],
    ],
)
def test_success(
    main_file: Union[IdentifiedJsonMain, IdentifiedPythonMain],
    labware_files: List[IdentifiedLabwareDefinition],
    data_files: List[IdentifiedData],
) -> None:
    """It should return everything in the input file list, separated by type."""
    subject = RoleAnalyzer()
    all_files = [main_file, *labware_files, *data_files]
    result = subject.analyze(all_files)
    assert result == RoleAnalysis(
        main_file=main_file,
        labware_files=labware_files,
        data_files=data_files,
    )


def test_error_if_no_files() -> None:
    """It should raise if the input file list contains no files."""
    subject = RoleAnalyzer()
    with pytest.raises(RoleAnalysisError, match="No files were provided."):
        subject.analyze([])


@pytest.mark.parametrize(
    "main_file_1, main_file_2",
    [
        # Python plus Python:
        (
            dummy_python_protocol_file(file_name="main_file_1"),
            dummy_python_protocol_file(file_name="main_file_2"),
        ),
        # Python plus JSON:
        (
            dummy_python_protocol_file(file_name="main_file_1"),
            dummy_json_protocol_file(file_name="main_file_2"),
        ),
        # JSON plus JSON:
        (
            dummy_json_protocol_file(file_name="main_file_1"),
            dummy_json_protocol_file(file_name="main_file_2"),
        ),
    ],
)
def test_error_if_multiple_main_files(
    main_file_1: IdentifiedFile, main_file_2: IdentifiedFile
) -> None:
    """It should raise if multiple main files are provided."""
    subject = RoleAnalyzer()
    with pytest.raises(
        RoleAnalysisError,
        match='Could not pick single main file from "main_file_1", "main_file_2"',
    ):
        subject.analyze([main_file_1, main_file_2])


@pytest.mark.parametrize(
    "labware_files, expected_message",
    [
        (
            [dummy_labware_definition_file(file_name="labware")],
            '"labware" is not a valid protocol file.',
        ),
        (
            [
                dummy_labware_definition_file(file_name="labware_1"),
                dummy_labware_definition_file(file_name="labware_2"),
            ],
            'No valid protocol file found in "labware_1", "labware_2"',
        ),
    ],
)
def test_error_if_files_but_no_main(
    labware_files: List[IdentifiedFile], expected_message: str
) -> None:
    """It should raise if no main files are provided, but other files are."""
    subject = RoleAnalyzer()
    with pytest.raises(
        RoleAnalysisError,
        match=expected_message,
    ):
        subject.analyze(labware_files)
